use pyo3::prelude::*;
use std::ops::Not;
use std::sync::Mutex;
use std::thread;

#[pyclass]
#[derive(Clone, Copy, PartialEq)]
enum Field {
    Empty,
    X,
    O,
}

const BOARD_SIZE: usize = 3;

#[pyclass]
#[derive(Clone, Copy)]
struct Game {
    #[pyo3(get, set)]
    board: [[Field; BOARD_SIZE]; BOARD_SIZE],
    x_moves: bool,
}

#[derive(PartialEq, Eq, PartialOrd, Ord, Clone, Copy)]
enum GameResult {
    Loss,
    Draw,
    Win,
}

impl Not for GameResult {
    type Output = Self;
    fn not(self) -> Self::Output {
        match self {
            Self::Draw => Self::Draw,
            Self::Win => Self::Loss,
            Self::Loss => Self::Win,
        }
    }
}

#[pyclass]
#[derive(Clone, Copy)]
struct Move(usize, usize);

impl Game {
    fn has_three_in_a_row(&self, field: Field) -> bool {
        self.board
            .into_iter()
            .any(|row| row.into_iter().all(|f| f == field))
            || (0..BOARD_SIZE)
                .into_iter()
                .any(|i| self.board.into_iter().all(|row| row[i] == field))
            || (0..BOARD_SIZE)
                .into_iter()
                .all(|i| self.board[i][i] == field)
            || (0..BOARD_SIZE)
                .into_iter()
                .all(|i| self.board[BOARD_SIZE - 1 - i][i] == field)
    }

    fn available_moves(&self) -> impl Iterator<Item = Move> + '_ {
        (0..BOARD_SIZE)
            .into_iter()
            .flat_map(|i| (0..BOARD_SIZE).into_iter().map(move |j| Move(i, j)))
            .filter(|&Move(i, j)| self.board[i][j] == Field::Empty)
    }

    fn exec_move_internal(mut self, Move(i, j): Move) -> Self {
        self.board[i][j] = if self.x_moves { Field::X } else { Field::O };
        self.x_moves = !self.x_moves;
        self
    }

    fn eval(mut self, mv: Move) -> GameResult {
        self = self.exec_move_internal(mv);

        if self.has_three_in_a_row(if self.x_moves { Field::X } else { Field::O }) {
            return GameResult::Win;
        }

        let mut moves = self.available_moves().peekable();
        if moves.peek().is_none() {
            return GameResult::Draw;
        }

        let mut best_result = GameResult::Loss;
        for mv in self.available_moves() {
            best_result = self.eval(mv).not().max(best_result);
            if let GameResult::Win = best_result {
                break;
            }
        }

        best_result
    }

    fn minmax_internal(self) -> Move {
        let bests = Mutex::new((Move(0, 0), GameResult::Loss));
        {
            let bests = &bests;

            thread::scope(|s| {
                for mv in self.available_moves() {
                    s.spawn(move || {
                        let result = !self.eval(mv);
                        let mut bests = bests.lock().unwrap();

                        if bests.1 <= result {
                            bests.0 = mv;
                            bests.1 = result;
                        }
                    });
                }
            });
        }

        bests.into_inner().unwrap().0
    }
}

#[pymethods]
impl Game {
    fn minmax(&self) -> (usize, usize) {
        let Move(i, j) = self.minmax_internal();
        (i, j)
    }

    fn exec_move(&mut self, mv: (usize, usize)) {
        self.board[mv.0][mv.1] = if self.x_moves { Field::X } else { Field::O };
        self.x_moves = !self.x_moves;
    }

    fn is_over(&self) -> bool {
        self.has_three_in_a_row(if self.x_moves { Field::O } else { Field::X })
            || self.available_moves().count() == 0
    }
}

#[pyfunction]
fn new_game() -> Game {
    Game {
        board: [[Field::Empty; BOARD_SIZE]; BOARD_SIZE],
        x_moves: true,
    }
}

#[pymodule]
fn zpr_bonus(_py: Python, m: &PyModule) -> PyResult<()> {
    m.add_function(wrap_pyfunction!(new_game, m)?)?;
    m.add_class::<Game>()?;
    Ok(())
}
